import numpy as np
from qssp.q_objects import measurement
from qssp.utils import *
import random

def test_init():
    ket3 = np.array([0.5, 0])
    rho3 = np.array([[0.25, 0],[0, 0]])
    state3 = qstate(ket3)
    np.testing.assert_allclose(state3.state, rho3)

def test_addition():
    ket_res = np.array([[1,0],[0,1]])
    np.testing.assert_allclose((ket0 + ket1).state, ket_res)

def test_substraction():
    ket_res = np.array([[0,1],[1,0]])
    np.testing.assert_allclose((ketp - ketm).state, ket_res)

def test_multiplication():
    ket_res = np.array([[2,0],[0,0]])
    np.testing.assert_allclose(2*ket0.state, ket_res)
    np.testing.assert_allclose((ket0 *2).state, ket_res)

def test_div():
    ket_res = np.array([[0,0],[0, 1/3]])
    np.testing.assert_allclose((ket1/3).state, ket_res)

def test_is_normalized():
    ket_norm = qstate(np.array([[1/2,0],[0,1/2]]))
    ket_not_norm = qstate(np.array([[0,3],[0,0]]))
    bloch_ket_norm = bloch_ket(np.pi)
    assert ket_norm.is_normalized() == True
    assert ket_not_norm.is_normalized() == False
    assert bloch_ket_norm.is_normalized() == True

def test_normalize():
    ket_res = np.array([[1/2,0],[0,1/2]])
    ket1.normalize()
    keta = ket0+ket1
    keta.normalize()
    np.testing.assert_allclose(keta.state, ket_res)

def test_is_hermitian():
    not_herm = qstate(np.array([[0, 1j],[0,0]]))
    assert ket0.is_hermitian()==True
    assert not_herm.is_hermitian() == False

def test_is_positive():
    neg_state = qstate(np.array([[2,0],[0,-1]]))
    assert ket1.is_positive() == True
    assert neg_state.is_positive() == False

def test_is_pure():
    max_mix = qstate(np.array([[1/2, 0],[0, 1/2]]))
    assert ketp.is_pure() == True
    assert max_mix.is_pure() == False

def test_is_valid():
    assert ket1.is_valid() == True
    assert (ket1 + ketp).is_valid() == False
    assert (qstate(np.array([[1,1],[0,0]]))).is_valid() == False
    assert (qstate(np.array([[2,0],[0,-1]]))).is_valid() == False

def test_vn_entropy():
    max_mix = qstate(np.array([[1/2, 0],[0, 1/2]]))
    assert ketm.vn_entropy() == 0
    assert max_mix.vn_entropy() == 1

def test_measure():
    np.testing.assert_allclose(ket0.measure(M_01), np.array([1, 0]))
    np.testing.assert_allclose(ket1.measure(M_01), np.array([0, 1]))
    np.testing.assert_allclose(ket0.measure(M_param(np.pi/2)), np.array([0.5, 0.5]))

def test_measure_sample():
    answ = np.array([1, 1, 1, 1, 0, 1, 0])
    np.random.seed(0)
    outcomes = ket0.measure_sample(M_param(np.pi/2), 7)
    assert np.all(outcomes == answ)

def test_m_labels():
    M01 = measurement([M_0, M_1], labels=['a','b'])
    #print (M01.labels)
    assert 'a' in M01.labels
    assert 'b' in M01.labels

def test_m_is_positive():
    M_neg = np.array([[-1,0],[0,0]])
    M_neg1 = measurement([M_neg, M_1])
    assert M_01.is_positive() == True
    assert M_neg1.is_positive() == False

def test_m_is_complete():
    M_inc = measurement([M_0])
    assert M_inc.is_complete() == False
    assert M_01.is_complete() == True

def test_r_mul():
    mul_state = 1/2 * ket0 + 2* ket1
    rmul_state = ket0 * 1/2 + ket1 * 2
    assert mul_state.state.all() == rmul_state.state.all()

def test_noise():
    np.testing.assert_allclose(ketp.add_noise('phaseflip',0.2).state,\
                                np.array([[0.5+0.j, 0.3+0.j],[0.3+0.j, 0.5+0.j]]))
    np.testing.assert_allclose(ket0.add_noise('bitflip',0.2).state,\
                                np.array([[0.8+0.j, 0.0+0.j],[0.0+0.j, 0.2+0.j]]))
    np.testing.assert_allclose((1/2*ket0+1/2*ketp).add_noise('bitphaseflip',0.2).state,\
                                np.array([[0.65+0.j, 0.15+0.j],[0.15+0.j, 0.35+0.j]]))
    np.testing.assert_allclose(ket0.add_noise('depolarizing',0.2).state,\
                                np.array([[0.9+0.j, 0.0+0.j],[0.0+0.j, 0.1+0.j]]))
    np.testing.assert_allclose(ket1.add_noise('amplitude_damping',0.2).state,\
                                np.array([[0.2+0.j, 0.0+0.j],[0.0+0.j, 0.8+0.j]]))


import numpy as np
from qssp.utils import GoldenMean, SNS, Nemo, RIP, Even

def test_GM_unifilar():
    assert GoldenMean().is_unifilar() == True

def test_SNS_unifilar():
    assert SNS().is_unifilar() == False

def test_Nemo_unifilar():
    assert Nemo().is_unifilar() == True

def test_Even_unifilar():
    assert Even().is_unifilar() == True

def test_RIP_unifilar():
    assert RIP().is_unifilar() == True

def test_GM_all_words():
    assert GoldenMean().all_words(3) == (['010','110','101','011','111'],[1/6,1/6,1/3,1/6,1/6])
    assert GoldenMean().all_words(0) == ([''], [1])

def test_GM_init_B_all_words():
    assert GoldenMean(init_dist=np.array([0,1])).all_words(3) == (['110','101','111'],[1/4,1/2,1/4])

def test_SNS_all_words():
    assert SNS().all_words(3) == (['000', '100', '010', '001', '101'],
 [0.3125, 0.18749999999999997, 0.25, 0.1875, 0.06249999999999999])

def test_GM_sample_words():
    all_words = GoldenMean().all_words(3)
    for word in GoldenMean().sample_words(5,3):
        assert word in all_words[0]

def test_SNS_sample_words():
    all_words = SNS().all_words(3)
    for word in SNS().sample_words(5,3):
        assert word in all_words[0]

def test_GM_stationary_dist():
    np.testing.assert_allclose(GoldenMean().stationary_distribution(), np.array([2/3,1/3]))

def test_SNS_stationary_dist():
    np.testing.assert_allclose(SNS().stationary_distribution(), np.array([1/2, 1/2]))

def test_GM_stationary_entropy():
    assert GoldenMean().state_entropy() == 0.9182958340544894
    assert GoldenMean().state_entropy([1,0]) == 0

def test_SNS_stationary_entropy():
    np.testing.assert_almost_equal(SNS().state_entropy(), 1)
    assert SNS().state_entropy([0,1])== 0

def test_GM_sample_transition():
    A_transitions = [(0,'1'),(1,'0')]
    B_transitions = [(0,'1')]
    sampled_A_transitions = [GoldenMean().sample_transition(0) for i in range(10)]
    sampled_B_transitions = [GoldenMean().sample_transition(1) for i in range(10)]

    for i in range(10):
        assert sampled_A_transitions[i] in A_transitions
        assert sampled_B_transitions[i] in B_transitions

def test_GM_block_entropies():
    np.testing.assert_allclose(GoldenMean().block_entropies(10), [0, 0.9182958340544894, 1.584962500721156, 2.251629167387823, 2.91829583405449, 3.584962500721157, 4.251629167387824, 4.91829583405449, 5.584962500721157, 6.251629167387822, 6.918295834054489])

def test_SNS_block_entropies():
    np.testing.assert_allclose(SNS().block_entropies(10), [0, 0.8112781244591328, 1.5, 2.1800365325772657, 2.8584585933443494, 3.5364882355820626, 4.2144073285182, 4.892292155061646, 5.570165618160434, 6.248035117839492, 6.925903179808621])

def test_GM_entropy_rate_approx():
    assert GoldenMean().entropy_rate_approx(1) == 0.9182958340544894
    assert GoldenMean().entropy_rate_approx(10) == 0.666666666666667

def test_SNS_entropy_rate_approx():
    assert SNS().entropy_rate_approx(1) == 0.8112781244591328
    assert SNS().entropy_rate_approx(10) == 0.6778680619691295

def test_GM_excess_entropy_approx():
    assert GoldenMean().excess_entropy_approx(1) == 0
    assert GoldenMean().excess_entropy_approx(5) == 0.25162916738782215
    assert GoldenMean().excess_entropy_approx(10) == 0.2516291673878195

def test_SNS_excess_entropy_approx():
    assert SNS().excess_entropy_approx(5) == 0.14634002439349647
    assert SNS().excess_entropy_approx(10) == 0.14722256011732604

def test_GM_evolve():
    GM_words = ['01', '10', '11']
    GM_evolutions_no_words = [np.array([[0.66666667, 0.33333333],
            [1.        , 0.        ]]),
        np.array([[0.66666667, 0.33333333],
            [0.        , 1.        ]]),
        np.array([[0.66666667, 0.33333333],
            [1.        , 0.        ]])]

    assert np.allclose(GM_evolutions_no_words[0], GoldenMean().evolve(2)) or np.allclose(GM_evolutions_no_words[1], GoldenMean().evolve(2)) or np.allclose(GM_evolutions_no_words[2], GoldenMean().evolve(2))
    
    path, word = GoldenMean().evolve(2,word=True)
    
    assert (np.allclose(path, GM_evolutions_no_words[0]) and word == GM_words[0]) \
           or (np.allclose(path, GM_evolutions_no_words[1]) and word == GM_words[1]) \
           or (np.allclose(path, GM_evolutions_no_words[2]) and word == GM_words[2])

def test_GM_many_paths():
    GM_mixed_states = np.array([[0.66666667, 0.33333333],
            [0.        , 1.        ],
            [1.        , 0.        ]])

    GM_path_states = GoldenMean().many_paths(2,10)
    
    for state in GM_path_states:
        assert np.allclose(GM_mixed_states[0], state) or np.allclose(GM_mixed_states[1], state) or np.allclose(GM_mixed_states[2], state)


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

#contains plotting functions that we use often

font = {'family': 'serif',
    'color':  'black',
    'weight': 'bold',
    'size': 35,
    'horizontalalignment' : 'right',
    'verticalalignment' : 'top'
    }

#Basic Mixed States Processing functions
####

def three_states_to_simplex(all_states):
    '''
    takes in an array of mixed states (from 3 state machine) and returns two ordered lists of x and y coordinates for the 2-simplex
    '''
    x = []
    y = []
    # run through keys and extract MSs
    for ms in range(len(all_states)):
        state = all_states[ms]
        a = state[0]; b = state[1]; c = state[2]
        # WLOG take two coordinates
        x.append(b+0.5*c)
        y.append(c*np.sqrt(3)/2.0)
    return x, y

def two_states_to_simplex(all_states):
    '''
    takes in an array of mixed states (from 2 state machine) and returns two ordered lists of x and y coordinates for the 1-simplex
    '''
    x = []
    for ms in xrange(len(all_states)):
        x.append(all_states[ms][0])
    return x

#Basic mixed states plotting functions
#####

def three_state_msp_scatter(x, y,filename='none'):
    '''
    takes in list of ordered x and y values and returns a scatter plot with mixed states in the 2-simplex
    '''
    plt.figure(figsize=(10,10))
    plt.clf()
    SimplexVertices = np.array([ [0,0],[1,0],[0.5,np.sqrt(3)/2.0],[0,0] ])
    plt.plot(SimplexVertices[:,0],SimplexVertices[:,1],color='k')
    plt.scatter(x, y, marker='.',alpha=0.1, s=2)

    plt.xticks([]); plt.yticks([])
    plt.xlim([0,1]); plt.ylim([0,1])
    plt.axis('off')
    plt.axis('equal')

    plt.text(0,-0.025,r'$(1,0,0)$',fontdict=font,ha='center',va='top')
    plt.text(1,-0.025,r'$(0,1,0)$',fontdict=font,ha='center',va='top')
    plt.text(0.5,np.sqrt(3)/2.0,r'$(0,0,1)$',fontdict=font,ha='center',va='bottom')
    if filename != 'none':
         #plt.savefig(filename+'.pdf',format='pdf',bbox_inches='tight')
        plt.savefig(filename+'.png',format='png',bbox_inches='tight')
    plt.show()
    return


#TODO: Add 2D plotting functionality: msp points in simplex, histogram
#TODO: Add 3D animation code


import numpy as np
import scipy.linalg as la
from scipy.stats import entropy

from .hmm import HMM
from .q_objects import qstate

def near(a, b, rtol = 1e-5, atol = 1e-8):
    return np.abs(a-b)<(atol+rtol*np.abs(b))

class qsHMM:
    '''
    A quantum state Hidden Markov Model consisting of a classical HMM
    and an alphabet of quantum states.
    
    Attributes
    -----------
    HMM : HMM object
        Numpy array of labeled transition matrices
    alph : list of qstates
        List of quantum states which are associated with symbols of the classical HMM
    alph_size : int
        Size of the alphabet of quantum states

    Methods
    -------
    q_word(word)
        Turns given classical word into quantum state using the quantum alphabet
    q_words(n, L)
        Samples words of length L with appropriate probabilities n times and returns list associated q_words
    q_block(L, join = True)
        Returns single quantum state represetning all words of length L
        or (if join == False) a quantum state and probability for each word
    q_block_entropies(L)
        Returns a list of von Neumann entropies for blocks of length=1 to length=L
    q_entropy_rate(L)
        Computes and returns the (difference) von Neumann entropy rate approximation at length L
    q_excess_entropy(L)
        Returns the quantum excess entropy approximation at length L
    get_measured_machine(measurement)
        Returns the classical HMM generator corresponding to a particular measurement basis
    observer_state_uncertainty(measurement, L, ever_synched = False)
        Returns average state uncertainty of observer measuring in fixed basis on up to L sites
    '''

    def __init__(self, model, alph, noise_type='None',noise_level=0):
        #should it take in Trans Matrices?
        if not isinstance(model, HMM):
            raise TypeError("Model must be specified as HMM object")
        self.HMM=model
        self.noise_type=noise_type
        self.noise_level=noise_level
        if noise_level != 0:
            self.alph = []
            for alph_state in alph:
                self.alph.append(alph_state.add_noise(noise_type, noise_level))
        else:
            self.alph= alph
        self.alph_size= int(len(alph))
        #check alph size is equal to no. of Ts
        
    def q_word(self, word):
        #takes in a classical word and turns it into a quantum state
        q_word = self.alph[int(word[0])].state
        for i in range(1,len(word)):
            q_word = np.kron(q_word, self.alph[int(word[i])].state)
        q_word = qstate(q_word)
        return q_word
        
    def q_words(self, n, L):
        #samples n words of length L and returns corresponding quantum states 
        if not isinstance(L, int):
            raise TypeError("Invalid word length (must be non-negative integer).")
        elif (L <= 0):
            raise ValueError("L must be greater than 0.")
        
        words = self.HMM.sample_words(n,L)
        q_words = [self.q_word(word) for word in words]
        return q_words
        
    def q_block(self, L, join=True):
        #calculates all the words of length L and their probs and returns the joint quantum state
        if not isinstance(L, int):
            raise TypeError("Invalid block length (must be non-negative integer).")
        elif (L <= 0):
            raise ValueError("L must greater than 0.")
        
        words, c_probs = self.HMM.all_words(L)
        if join:
            q_block = c_probs[0] * self.q_word(words[0]).state
            for i in range(1,len(words)):
                q_block = q_block + c_probs[i]*self.q_word(words[i]).state
            q_block = qstate(q_block)
            return q_block
        else: 
            q_seq = []
            for i in range(len(words)):
                q_seq.append(self.q_word(words[i]).state)
            return q_seq, c_probs

    def q_block_entropies(self, L):
        if not isinstance(L, int):
            raise TypeError("Invalid block length (must be non-negative integer).")
        elif (L <= 0):
            raise ValueError("L must be greater than 0.")
        
        ents = [0] + [self.q_block(l).vn_entropy() for l in range(1,L+1)]
        return ents
        
    def q_entropy_rate(self, L):
        if not isinstance(L, int):
            raise TypeError("Invalid block length (must be non-negative integer).")
        elif (L < 1):
            raise ValueError("L must be greater than 0.")
        
        q_ents = self.q_block_entropies(L)
        s_est = q_ents[-1] - q_ents[-2]
        return s_est
        
    def q_excess_entropy(self, L):
        if not isinstance(L, int):
            raise TypeError("Invalid block length (must be non-negative integer).")
        elif (L < 1):
            raise ValueError("L must be greater than 0.")
        
        EE_est = self.q_block(L).vn_entropy() - L * self.q_entropy_rate(L)
        return EE_est

    def get_measured_machine(self, measurement):
        '''
        Computes the measured machine when applying "measurement" to the qs_HMM, returns an HMM object
        '''
        #size of the classical alphabet of the measured machine
        c_abet_size = len(measurement.mOps)
        #number of states in the HMM
        n_states = self.HMM.Ts[0].shape[0]

        #initialize an array that will store the labeled transition matrices for the measured machine
        measured_Ts = np.zeros([c_abet_size, n_states, n_states])
        
        #each iteration over i computes th labeled transition matrix of the measured machine T_i
        for i in range(c_abet_size):
            #Initialize labeled transition matrix
            t = np.zeros([n_states, n_states])
            for j in range(self.alph_size):
                #compute the probability of measurement outcome i when measuring quantum state j
                prob = self.alph[j].measure(measurement)[i]
                #weigh the quantum labeled transition matrix of state j by the probability of seeing out come i, and add that for all js
                t += (self.HMM.Ts[j]*prob)
            #store labeled transition matrix T_i    
            measured_Ts[i] = t
        #create HMM object with the correct transition matrices
        measured_mach = HMM(measured_Ts)
        return measured_mach

    def observer_state_uncertainty(self,measurement,L,ever_synched=False):
        '''
        Returns average state uncertainty of observer measuring in fixed basis on up to L sites
        '''

        S = [[self.HMM.init]]
        probs = [[1]]
        H_avg = [entropy(S[0][0],base=2)]
        
        meas = self.get_measured_machine(measurement)
        Ts = meas.Ts
        
        for l in range(1,L+1):
            S_t = []
            probs_t = []
            H_avg_t = 0

            for s in range(len(S[-1])):
                # Calculates mixed states given L observations using mixed states given (L-1) observations
                S_prev = S[-1][s]
                prob_prev = probs[-1][s]

                if(ever_synched and self.is_synched(S_prev)):
                    # If 'ever_synched', stop calculating new mixed states after synchronizing observation
                    S_t.append(S_prev)
                    probs_t.append(prob_prev)
                
                else:
                    for i in range(len(meas.Ts)):
                        S_new = [0] * len(meas.Ts[0])

                        for k in range(len(meas.Ts[0])):
                            prob_k = 0

                            for j in range(len(meas.Ts[0])):
                                if not near(Ts[i][j][k],0):
                                    # Calculate probability of each new mixed state
                                    prob_k += S_prev[j]* Ts[i][j][k]

                            S_new[k] += prob_k

                        if not near(sum(S_new),0):
                            S_t.append(S_new/sum(S_new))
                            new_prob = prob_prev*sum(S_new)
                            probs_t.append(new_prob)
                            H_avg_t += new_prob * entropy(S_new/sum(S_new),base=2)

            S.append(S_t)
            probs.append(probs_t)
            H_avg.append(H_avg_t)

        # Returns list of entropies over mixed states for different L values
        return H_avg

    def is_synched(self, mixed_state):
        # If mixed state is concentrated in one generator state returns True, else False
        if not isinstance(mixed_state, list):
            raise TypeError("Mixed state must be list of probabilities")
        elif not len(mixed_state) == len(self.HMM.stationary_distribution()):
            raise ValueError("Mixed state has the wrong number of elements.")
        elif not near(sum(mixed_state),1):
            raise ValueError("Mixed state probabilities must sum to 1.")

        for s in mixed_state:
            if (not near(s,0) and not near(s,1)):
                return False
        
        return True


import numpy as np
import scipy.linalg as la
from scipy.stats import entropy

def near(a, b, rtol = 1e-5, atol = 1e-8):
    return np.abs(a-b)<(atol+rtol*np.abs(b))

def expect(mOp, state):
    #expectation value of measuring state with mop
    return np.real(np.trace(state @ mOp))

class qstate:
    """A discrete state in a Hilbert space
    
    Attributes
    ----------
    dim : int
        dimension of the Hilbert space
    state: complex numpy array
        density matrix describing the quantum state

    Methods
    -------
    is_normalized()
        Returns True if state is normalized array, False otherwise
    normalize()
        Normalizes state if not already normalized
    is_hermitian()
        Returns True if state is Hermitian array
    is_positive()
        Returns True if all eigenvalues of state matrix are 0 or positive, False otherwise
    is_valid()
        Returns True if state is normalized, Hermitian and postive, False otherwise
    is_pure()
        Returns True if state matrix is approximately idempotent
    vn_entropy(base=2)
        Returns von Neumann entropy of state matrix with desired base
    measure(measurement, labels=False)
        Applies measurement to state and returns probabilities of different measurement outcomes
    measure_sample(measurement, num_samples)
        Returns measurement outcomes when applying measurement sampled 'num_samples' times with correct probabilities
    """

    def __init__(self, amplitudes, test_norm = True):
        """
        Parameters
        ----------
        amplitudes: numpy array
            Array describing quantum state as vector (pure) or matrix (mixed)
        """

        amplitudes = amplitudes.astype(complex)
        
        #if amplitudes is ket --> dm
        if amplitudes.ndim == 1:
            self.dim = int(len(amplitudes))
            self.state = np.outer(amplitudes.conj().T,amplitudes)
            if(test_norm and (not self.is_normalized())):
                print("WARNING: Not normalized")
        
        #if amplitudes is matrix -> dm
        elif amplitudes.ndim >= 2:
            self.dim = int(len(amplitudes[0]))
            self.state = amplitudes
            if(test_norm and (not self.is_normalized())):
                print("WARNING: Not normalized")
        
        else:
            raise ValueError("Invalid input dimensions")
    
    def __add__(self, state_2):
        """ Overloaded addition for qstates."""
        sum = self.state + state_2.state

        return qstate(np.array(sum))

    def __sub__(self, state_2):
        """ Overloaded subtraction for qstates."""
        diff = self.state - state_2.state

        return qstate(np.array(diff))

    def __mul__(self, num):
        """ Overloaded multiplication for qstates."""
        mult = num * self.state

        return qstate(np.array(mult), test_norm=False)

    def __rmul__(self, num):
        """ Overloaded reverse multiplication for qstates."""
        mult = self.state * num

        return qstate(np.array(mult), test_norm=False)

    def __truediv__(self, num):
        """ Overloaded division for qstates."""
        div = self.state / num

        return qstate(np.array(div), test_norm=False)

    def is_normalized(self):
        """ Returns True if state is normalized array, False otherwise."""

        if near(np.trace(self.state),1):
            return True
        else:
            return False
        
    def normalize(self):
        """Normalizes state if not already normalized."""

        if self.is_normalized():
            pass
        else:
            norm = np.trace(self.state)
            self.state = self.state/norm
            pass
    
    def is_hermitian(self):
        """Returns True if state is Hermitian array."""

        return np.allclose(self.state, np.conj(self.state.T))
    
    def is_positive(self):
        """Returns True if all eigenvalues of state matrix are 0 or positive, False otherwise."""
       
        eigs = la.eig(self.state)[0]
        
        return not any(not near(eig,0) and eig < 0 for eig in eigs)
        
    def is_valid(self):
        """Returns True if state is normalized, Hermitian and postive, False otherwise."""
        
        normalized = self.is_normalized()
        hermitian = self.is_hermitian()
        positive = self.is_positive()
        
        if (normalized & hermitian & positive):
            return True
        else:
            if not normalized:
                print('Not Normalized')
            if not hermitian:
                print('Not Hermitian')
            if not positive:
                print('Not Positive')
            return False
        
    def is_pure(self):
        """Returns True if state matrix is approximately idempotent."""
        sq_trace = np.trace(np.matmul(self.state,self.state))
        is_p = near(sq_trace,1)
        return is_p
##NOTE: DELETED THE FOLLOWING CODE, IT WAS INCORRECT. DO WE WANT A PRINT OPTION?
       # if self.is_normalized():
        #    return (near(np.trace(np.dot(self.state,self.state)),1))
       # else:
       #     print('Not Pure')
       #     return False
        
    def vn_entropy(self,base=2):
        """Returns von Neumann entropy of state matrix with desired base.
        
        Parameters
        ----------
        base : real
            Base of logarithm to use (default is 2)
        """
        
        eigs = la.eig(self.state)[0]
        eigs = [0 if near(e,0) else e.real for e in eigs]
        return entropy(eigs,base=base)
    
    def measure(self, measurement, labels=False):
        """Applies measurement to state and returns probabilities of different measurement outcomes.
        
        Parameters
        ----------
        measurement : measurement class
            Measurement operators to apply to quantum state
        labels : Boolean
            If True, returns list of strings labeling measurement outcomes, default is False
        """
        
        probs = np.array([expect(mop, self.state) for mop in measurement.mOps])
        return probs
    
    def measure_sample(self, measurement, num_samples):
        """Returns measurement outcomes sampled with the correct probabilities.

        Parameters
        ----------
        measurement: instance of measurement class
        """

        outcomes = np.zeros(num_samples)
        for i in range(num_samples):
            outcome = np.random.choice(measurement.labels, p = self.measure(measurement))
            outcomes[i]=outcome
        return outcomes

    def add_noise(self,noise_type,noise_level):
        """Returns state after passing it through a noisy channel.

        Parameters
        ----------
        noise_type: str
            Type of noisy channel to use, can be 'phaseflip', 'bitflip',
            'bitphaseflip', 'depolarizing', or 'amplitude_damping'
        noise_level: float
            Amount of noise to apply
        """

        if noise_type == 'phaseflip':
            E_0 = np.sqrt(1-noise_level)*np.array(([1,0],[0,1]))
            E_1 = np.sqrt(noise_level)*np.array(([1,0],[0,-1]))
            noisy_state = E_0 @ self.state @ (E_0.conj().T) + E_1 @ self.state @ (E_1.conj().T) 
        elif noise_type == 'bitflip':
            E_0 = np.sqrt(1-noise_level)*np.array(([1,0],[0,1]))
            E_1 = np.sqrt(noise_level)*np.array(([0,1],[1,0]))
            noisy_state = E_0 @ self.state @ (E_0.conj().T) + E_1 @ self.state @ (E_1.conj().T) 
        elif noise_type == 'bitphaseflip':
            E_0 = np.sqrt(1-noise_level)*np.array(([1,0],[0,1]))
            E_1 = np.sqrt(noise_level)*np.array(([0,-1j],[1j,0]))
            noisy_state = E_0 @ self.state @ (E_0.conj().T) + E_1 @ self.state @ (E_1.conj().T)
        elif noise_type == 'depolarizing':
            noisy_state = noise_level/2 * np.array([[1,0],[0,1]]) + (1-noise_level) * self.state
        elif noise_type == 'amplitude_damping':
            E_0 = np.array(([1,0],[0,np.sqrt(1-noise_level)]))
            E_1 = np.array(([0,np.sqrt(noise_level)],[0,0]))
            noisy_state = E_0 @ self.state @ (E_0.conj().T) + E_1 @ self.state @ (E_1.conj().T)
        else:
            raise ValueError("Unknown noise type: %s" % noise_type)

        return(qstate(noisy_state))
    
class measurement:
    """A quantum measurement, consisting of a set of operators
    
    Attributes
    ----------
    mOps : list of numpy arrays 
        measurement operators in matrix form
    n_ops: int
        number of measurement operators
    tol_positvity: float
        numerical tolerance for checking positivity/completeness of measurement

    Methods
    -------
    is_postive()
        Returns True if all eigenvalues of measurement operators are non-negative.
    is_complete()
        Returns True if measurement operators sum to the identity (within tolerance).
    """

    def __init__(self, mOps, labels=None, tol_positivity=1e-8):
        """
        Parameters
        ----------
        mOps: list of numpy arrays
            Measurement operators in matrix form
        labels: list of strings
            List of labels corresponding to measurement outcomes, default is None
        tol_postivity: float
            Tolerance for checking completeness/positivity of measurement, default is 1e-8
        """

        #TODO: check sum to identity, check positivity
        self.mOps = mOps
        self.n_ops = int(len(self.mOps))
        if labels==None:
            self.labels = range(self.n_ops)
        else:
            self.labels=labels
        self.tol_positivity= tol_positivity

    def is_positive(self):
        """Returns True if all eigenvalues of measurement operators are non-negative, otherwise returns False."""

        for mop in self.mOps:
            vs = la.eigvals(mop)
            for v in vs:
                #assert val.imag<tol_positivity
                #assert val.real >= -1*tol_positivity
                a = v.imag<self.tol_positivity
                b = v.real >= -1*self.tol_positivity
                if (a&b):
                    pass
                else:
                    return False
        return True

    def is_complete(self):
        """Returns True if measurement operators sum to the identity (within tolerance), otherwise returns False."""
        
        return np.allclose(np.sum(self.mOps,axis=0),np.identity(self.n_ops))
    
    #TODO: is_valid


import numpy as np
from .hmm import HMM
from .q_objects import qstate, measurement
# Some useful Machines 

#Two state machines

def GoldenMean(p=0.5,init_dist=None): 
    t0 = np.array([[0, p],[0,0]])
    t1 = np.array([[1-p, 0],[1,0]])
    return HMM([t0,t1],init_dist)

def Even(p=0.5,init_dist=None):
    t0 = np.array([[1-p, 0],[0,0]])
    t1 = np.array([[0, p],[1,0]])
    return HMM([t0,t1],init_dist)

def SNS(p=0.5,q=0.5,init_dist=None):
    t0 = np.array([[p, 1-p],[0,q]])
    t1 = np.array([[0, 0],[1-q,0]])
    return HMM([t0,t1],init_dist)


#Three State Machines
def Nemo(p=0.5, q=0.5,init_dist=None):
    t0 = np.array([[p,0,0],[0,0,0],[q,0,0]])
    t1 = np.array([[0,1-p,0],[0,0,1],[1-q,0,0]])
    return HMM([t0,t1],init_dist)

def RIP(p=0.5,q=0.5,init_dist=None):
    t0 = np.array([[0,p,0],[0,0,q],[0,0,0]])
    t1 = np.array([[0,0,1-p],[0,0,1-q],[1,0,0]])
    return HMM([t0,t1],init_dist)


# #Some useful states 
identity = qstate(np.array([[1,0],[0,1]]), test_norm=False)
sigma_x = qstate(np.array([[0,1],[1,0]]), test_norm=False)
sigma_y = qstate(np.array([[0,-1j],[1j,0]]), test_norm=False)
sigma_z = qstate(np.array([[1,0],[0,-1]]), test_norm=False)

ket0a = np.array([1,0])
ket1a = np.array([0,1])
ketpa = (1/np.sqrt(2))*(ket0a+ket1a)
ketma = (1/np.sqrt(2))*(ket0a-ket1a)

def bloch_ket(theta, phi=0):
    a = np.cos(theta/2)
    b = np.sin(theta/2)
    ket = a*ket0a + b*np.exp(1j*phi)*ket1a
    return qstate(ket)

ket0 = qstate(ket0a)
ket1 = qstate(ket1a)
ketp = qstate(ketpa)
ketm = qstate(ketma)

A_01 = np.array([ket0,ket1])
A_0p = np.array([ket0,ketp])
A_pm = np.array([ketp,ketm])

M_0 = np.array([[1,0],[0,0]])
M_1 = np.array([[0,0],[0,1]])
M_01 = measurement([M_0,M_1])

def M_param(theta, phi=0):
    a = np.cos(theta/2)
    b = np.sin(theta/2)
    ket = a*ket0a + b*np.exp(1j*phi)*ket1a
    M0 = np.outer(ket.conj().T, ket)
    M1 = np.identity(2) - M0
    return measurement([M0,M1])


import numpy as np
import scipy.linalg as la
from scipy.stats import entropy

def near(a, b, rtol = 1e-5, atol = 1e-8):
    return np.abs(a-b)<(atol+rtol*np.abs(b))

class HMM:
    '''
    A Hidden Markov Model representation of a process with basic functionality. 
    
    Attributes
    -----------
    Ts : np.array
        Numpy array of labeled transition matrices
    init_dist : np.array
        Array containing the initial probability distribution over hidden states. If None, the stationary state distribution is used
    alphabet : np.array
        Array containing symbols alphabet for HMM
    state_labels : np.array
        Array containing labels for states of HMM

    Methods
    -------
    is_unifilar()
        Bool- returns True if HMM is unifilar
    all_words(L)
        Returns all possible words of length L and their probabilities
    sample_transition(state)
        Samples a possible transition from a given hidden state, with its associated probability. Returnsnext state and the emitted symbol
    sample_words(n, L)
        Samples and returns n words of length L
    stationary_distribution()
        Returns the stationary state distribution of the HMM as a numpy array
    state_entropy(dist=None)
        Returns the Shannon entropy of the given state distribution dist. If None, returns the entropy of the stationary state distribution
    block_entropies(L)
        Returns a list of block entropies for words of length=1 to length=L
    entropy_rate_approx(L)
        Computes and returns the (difference) entropy rate approximation at length L
    excess_entropy_approx(L)
        Returns the Excess Entropy approximation at length L
    '''
    def __init__(self, Ts, init_dist = None):
        '''
        Parameters
        -----------
        Ts: np.array
            Labeled transition matrices of the HMM
        init_dist=None: None or np.array
            Initial state distribution. If none passed, it is taken to be the asymptotic state distribution. 
        '''
        self.Ts = Ts
        self.alphabet  = np.arange(len(Ts))
        self.state_labels = np.arange(len(Ts[0]))
        
        #initialize state distribution as pi if not specified
        if init_dist is None:
            self.init_dist = self.stationary_distribution()
        else:
            self.init_dist = init_dist

    def is_unifilar(self):
        '''bool - returns True if HMM is unifilar'''
        for T in self.Ts:
            #for each labelled transition matrix count number of non zero entries per row and flag false if any is bigger than 0
            non_zero= np.count_nonzero(T, axis=1)
            count= len(non_zero[non_zero>1])
            if count > 0:
                return False

        return True

    def all_words(self, L):
        '''Returns all possible words of length L and their probabilities
        Parameters
        ----------
        L: int
            Length of words to be computed
        Returns
        -------
        tuple
            (words, probs) are lists of allowed words and their associated probabilities
        '''
        init_dist = self.init_dist
        Ts = self.Ts

        if L < 1:
            words = ['']
            probs = [1]
            
            return words, probs

        else:

            words = [str(x) for x in range(len(Ts))]
            probs = []

            for l in range(1,L):
                words_l = []
                for x in range(len(Ts)):
                    for word in words:
                        words_l.append(word + str(x))
                words = words_l

            for word in list(words):
                prob = init_dist.T
                for x in word:
                    prob = np.matmul(prob,Ts[int(x)])
                prob = sum(prob)
                if near(prob,0):
                    words.remove(word)
                else:
                    probs.append(prob)

            return words, probs

    def sample_transition(self, state):
        ''' Samples a possible transition from a given hidden state, with its associated probability. Returnsnext state and the emitted symbol
        Parameters
        -----------
        state: int
            State from which a transition will be sampled
        Returns
        -------
        tuple
            (new_state, symbol) are both int representing the new state and the symbol emitted during transition
        '''
        transition_states = []
        transition_symbols = []
        probs = []

        Ts = self.Ts

        for x in range(len(Ts)):
            for j in range(len(Ts[x][state])):
                #probability of transitioning to state j on symbol x
                prob = Ts[x][state][j]
                if prob > 0:
                    transition_states.append(j)
                    transition_symbols.append(str(x))
                    probs.append(prob)

        #select a transition from the list of all possible transitions based upon probabilities
        new_index = np.random.choice(np.arange(0,len(probs)),p=probs)
        new_state = transition_states[new_index]
        symbol = transition_symbols[new_index]

        return new_state, symbol

    def sample_words(self, n, L):
        '''
        Samples and returns n words of length L
        Parameters
        ----------
        n: int
            Number of words to be sampled and returned
        L: int
            Length of words to be sampled and returned
        Returns
        -------
        list 
            List of sampled words
        '''
        words = []

        #determine initial state for all n words
        init_dist = np.random.choice(np.arange(0,len(self.init_dist)),n,p=self.init_dist)

        for i in range(n):
            word_n = ''

            # set initial state vector
            state = init_dist[i]

            for l in range(L):

                # randomly (with weighted probabilities) choose a transition from current state
                state, symbol = self.sample_transition(state)
                word_n = word_n + symbol

            words.append(word_n)
        return words

    def stationary_distribution(self):
        '''Returns the stationary state distribution of the HMM as a numpy array'''
        #transition matrix of the HMM
        t = np.sum(self.Ts,0)
        t = t.T
        #compute eigenvalues and eigenvector
        d, v = la.eig(t)
        #select eigv with eigv 1
        pi = v[:, near(d,1.0)].T[0]
        #normalize
        pi = np.real(pi/np.sum(pi))
        return pi.T

    def state_entropy(self, dist=None):
        '''Returns the Shannon entropy of the given state distribution dist. If None, returns the entropy of the stationary state distribution
        Parameters
        ----------
        dist=None : dist
            Distribution whose Shannon entropy will be computed, if None then returns Shannon Entropy of stationary state distribution
        Returns
        -------
        float
            Shannon entropy rate of state distribution
         '''
        #state entropy of the stationary distribution
        if dist is None:
            dist = self.stationary_distribution()
        H = entropy(dist, base=2)
        return H

    def block_entropies(self, L):
        '''Returns a list of block entropies for words of length=0 to length=L
        Parameters
        ----------
        L: int
            Maximum word length for which block entropies will be computed
        Returns
        -------
        list
            List of block entropies for words from length 1 to L
        '''
        block_entropies = [0]

        for l in range(1,L+1):
            block_entropies.append(entropy(self.all_words(l)[1],base=2))

        return block_entropies

    def entropy_rate_approx(self, L):
        '''
        Computes and returns the (difference) entropy rate approximation at length L
        Parameters
        ----------
        L: int
            Length for which entropy rate approximation is computed.
        Returns
        -------
        float
            Entropy rate approximation
        '''

        hmu_L = entropy(self.all_words(L)[1],base=2) - entropy(self.all_words(L-1)[1],base=2)

        return hmu_L

    def excess_entropy_approx(self, L):
        ''' Returns the Excess Entropy approximation at length L
        Parameters
        ----------
        L: int
            Wordlength at which the entropy rate approximation is computed
        Returns
        -------
        float
            Excess entropy approximation
        '''
        block_entropies = self.block_entropies(L)
        #print(L, block_entropies)

        if L > 1:
            EE_L = block_entropies[-1] - L * (block_entropies[-1] - block_entropies[-2])
        else: 
            EE_L = 0

        return EE_L


    def evolve(self, N, init_dist=None, transients=0, word=False):
        '''
        takes an initial mixed state and evolves it N time steps (a single path)

        Parameters
        -----------
        N: int
            Number of time steps to be computed
        init_dist=None: None or np.array
            Initial mixed state
        transients=0: int
            Number of transients to be thrown away from the path.
        word: bool, optional
            If True returns word generated during evolution 

        Returns
        --------
        np.array with the mixed states visited in the path, will have N-transients entries
        '''

        ts = self.Ts
        num_states = len(ts[0])
        mstates = np.zeros((N, num_states))

        if init_dist is None:
            init_dist = self.init_dist

        long_word = self.sample_words(1, N)[0]
        mstates[0]=init_dist

        for i in range(1,len(long_word)):
            mu = np.matmul(mstates[i-1], ts[int(long_word[i])])
            mu = mu/np.sum(mu)
            if not near(np.sum(mu), 1.0):
                raise ValueError("Normalization failed for length-" + str(i) + " mixed state" )
            mstates[i]=mu
        
        mstates = mstates[transients:]
        if word:
            return mstates, long_word
        else:
            return mstates

    def many_paths(self, N, runs, init_dist=None, transients=0):
        #TODO: allow different initial states for every run?
        '''
        Runs evolve 'runs' times and stitches all the resulting mixed states into one array. 
        For any well behaved hmm and a sufficiently large number of N and runs, this should be 
        an approximation of the msp states (w/o transitions)

        Parameters:
        -----------
        N: int
            number of timesteps to evolve initial state
        runs: int
            number of times the evolution is run (starting in the same initial state)
        init_dist: np.array or None
            initial state of hmm
        transients: int
            number of transient states to be ignored for each run of the evolution

        Returns:
        --------
        np.array with the mixed states visited during 'runs' runs of length (N-transients)
        '''
        ms_all = self.evolve(N, init_dist, transients)
        for i in range(runs-1):
            path = self.evolve(N, init_dist, transients)
            ms_all = np.concatenate((ms_all, path),0)
        
        return ms_all